from decimal import Decimal
from ameisedataset.miscellaneous import obj_to_bytes, obj_from_bytes, read_data_block
from ameisedataset.data import Tower, Vehicle, VisionSensorsVeh, VisionSensorsTow, LaserSensorsVeh, LaserSensorsTow


class Frame:
    def __init__(self, frame_id: int, timestamp: Decimal):
        """
        Initializes the Frame object with the provided frame ID and timestamp.
        Sets default values for cameras and lidar attributes.
        Parameters:
            frame_id (int): Unique identifier for the frame.
            timestamp (Decimal): Timestamp associated with the frame.
        """
        self.frame_id: int = frame_id
        self.timestamp: Decimal = timestamp
        self.vehicle: Vehicle = Vehicle()
        self.tower: Tower = Tower()

    def to_bytes(self) -> bytes:
        # dump meta info like timestamp
        meta_bytes = obj_to_bytes([self.frame_id, self.timestamp])
        # vehicle bytes
        veh_bytes = self.vehicle.to_bytes()
        # tower bytes
        tow_bytes = self.tower.to_bytes()
        # always dumped in same pattern: len + content
        frame_bytes = meta_bytes + veh_bytes + tow_bytes
        return frame_bytes

    @classmethod
    def from_bytes(cls, data: bytes) -> "Frame":
        # read block (len + content) by block and divide into fixed sequence: meta, vehicle, tower
        meta_bytes, data = read_data_block(data)
        vehicle_bytes, data = read_data_block(data)
        tower_bytes, data = read_data_block(data)
        # load meta
        meta_data = obj_from_bytes(meta_bytes)
        # create object with metadata
        frame = cls(frame_id=meta_data[0], timestamp=meta_data[1])
        # vehicle content
        frame.vehicle = Vehicle.from_bytes(vehicle_bytes)
        # tower content
        frame.tower = Tower.from_bytes(tower_bytes)
        return frame

    def is_complete(self):
        """ Checks if all fields of the frame object are filled and returns a list of unfilled fields.
        Returns:
            list: List of strings representing the unfilled fields.
        """
        unfilled_fields = []
        for agent in [self.vehicle, self.tower]:
            for attr_name, attr_value in agent.__dict__.items():
                if isinstance(attr_value, (VisionSensorsVeh, VisionSensorsTow, LaserSensorsVeh, LaserSensorsTow)):
                    for sub_attr_name, sub_attr_value in attr_value.__dict__.items():
                        if sub_attr_value is None:
                            unfilled_fields.append(sub_attr_name)
                elif attr_value is None:
                    # check for len(obj.attr)
                    unfilled_fields.append(attr_value)
        return True if not unfilled_fields else unfilled_fields



import numpy as np
from PIL import Image as PilImage
from ameisedataset.data import CameraInformation, Image
import cv2

def rectify_image(image: Image, camera_information: CameraInformation):
    """Rectify the provided image using camera information."""
    # Init and calculate rectification matrix
    mapx, mapy = cv2.initUndistortRectifyMap(cameraMatrix=camera_information.camera_mtx,
                                             distCoeffs=camera_information.distortion_mtx[:-1],
                                             R=camera_information.rectification_mtx,
                                             newCameraMatrix=camera_information.projection_mtx,
                                             size=camera_information.shape,
                                             m1type=cv2.CV_16SC2)
    # Apply matrix
    rectified_image = cv2.remap(np.array(image.image), mapx, mapy, interpolation=cv2.INTER_LINEAR)

    return Image(PilImage.fromarray(rectified_image), image.timestamp)


def create_stereo_image(image_left: Image, image_right: Image, cam_right_info: CameraInformation) -> np.ndarray:
    """
    Create a disparity map from two rectified images.
    Parameters:
    - image_left: First image as a PIL Image.
    - image_right: Second image as a PIL Image.
    - cam_right_info: Camera Info object
    Returns:
    - Depth map as a numpy array.
    """
    # Convert PIL images to numpy arrays
    img1 = np.array(image_left.convert('L'))  # Convert to grayscale
    img2 = np.array(image_right.convert('L'))  # Convert to grayscale
    # Create the block matching algorithm with high-quality settings
    stereo = cv2.StereoSGBM_create(
        minDisparity=0,
        numDisparities=128,         # Depending on the camera setup, this might need to be increased.
        blockSize=5,                # Smaller block size can detect finer details.
        P1=8 * 3 * 5 ** 2,          # Control smoothness of the disparity. Adjust as needed.
        P2=32 * 3 * 5 ** 2,         # Control smoothness. This is usually larger than P1.
        disp12MaxDiff=1,            # Controls maximum allowed difference in disparity check.
        uniquenessRatio=15,         # Controls uniqueness. Higher can mean more robustness against noise.
        speckleWindowSize=100,
        speckleRange=32,
        mode=cv2.STEREO_SGBM_MODE_SGBM_3WAY  # Utilizes 3-way dynamic programming. May provide more robust results.
    )
    # Compute the disparity map
    disparity = stereo.compute(img1, img2)
    # Normalize for better visualization
    disparity = cv2.normalize(disparity, disparity, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    # To avoid division by zero, set disparity values of 0 to a small value
    safe_disparity = np.where(disparity == 0, 0.000001, disparity)
    f = cam_right_info.focal_length
    b = abs(cam_right_info.stereo_transform.translation[0]) * 10 ** 3
    depth_map = f * b / safe_disparity
    return depth_map



class InfoBaseClass(type):
    """ Base metaclass to provide dictionary-like access to class attributes.
    Allows retrieval of class attributes using dictionary key access style.
    """
    def __getitem__(cls, key):
        """ Retrieve class attribute value using key access.
        Args:
            key (str): Attribute key.
        Returns:
            Any: Value of the specified class attribute.
        """
        return getattr(cls, key)

    def is_type_of(cls, value):
        """ Check if the provided value corresponds to a defined camera type.
        Args:
            value (Any): Value to be checked.
        Returns:
            bool: True if the value is a defined camera type, False otherwise.
        """
        return value in cls.__dict__

    def get_name_by_value(cls, value):
        """ Retrieve the name of the class constant based on its value. """
        constants = {k: v for k, v in vars(cls).items() if not callable(v) and not k.startswith("__")}
        for name, val in constants.items():
            if val == value:
                return name
        return None


class Camera(metaclass=InfoBaseClass):
    """ Defines constants representing different camera types.
    Attributes:
        MONO_LEFT (int): Represents a mono left camera.
        STEREO_LEFT (int): Represents a stereo left camera.
        STEREO_RIGHT (int): Represents a stereo right camera.
        MONO_RIGHT (int): Represents a mono right camera.
    """
    BACK_LEFT = 0
    FRONT_LEFT = 1
    STEREO_LEFT = 2
    STEREO_RIGHT = 3
    FRONT_RIGHT = 4
    BACK_RIGHT = 5
    REAR = 6
    VIEW_1 = 7
    VIEW_2 = 8


class Lidar(metaclass=InfoBaseClass):
    """ Defines constants representing different Lidar types.
    Attributes:
        OS0_LEFT (int): Represents a OS0 left Lidar.
        OS1_TOP (int): Represents a OS1 top Lidar.
        OS0_RIGHT (int): Represents a OS0 right Lidar.
    """
    LEFT = 0
    TOP = 1
    RIGHT = 2


class IMU(metaclass=InfoBaseClass):
    BUS = 0


class GNSS(metaclass=InfoBaseClass):
    BUS = 0
    TOWER = 1



from typing import Optional
from ameisedataset.data import Camera, Lidar, IMU, GNSS, Dynamics
from ameisedataset.miscellaneous import serialize, deserialize, INT_LENGTH


class VisionSensorsVeh:
    def __init__(self):
        self.BACK_LEFT: Optional[Camera] = None
        self.FRONT_LEFT: Optional[Camera] = None
        self.STEREO_LEFT: Optional[Camera] = None
        self.STEREO_RIGHT: Optional[Camera] = None
        self.FRONT_RIGHT: Optional[Camera] = None
        self.BACK_RIGHT: Optional[Camera] = None
        self.REAR: Optional[Camera] = None

    def to_bytes(self):
        return b''.join(serialize(camera) for camera in [
            self.BACK_LEFT,
            self.FRONT_LEFT,
            self.STEREO_LEFT,
            self.STEREO_RIGHT,
            self.FRONT_RIGHT,
            self.BACK_RIGHT,
            self.REAR
        ])

    @classmethod
    def from_bytes(cls, data) -> 'VisionSensorsVeh':
        instance = cls()
        for attr in ['BACK_LEFT', 'FRONT_LEFT', 'STEREO_LEFT', 'STEREO_RIGHT', 'FRONT_RIGHT', 'BACK_RIGHT', 'REAR']:
            camera, data = deserialize(data, Camera)
            setattr(instance, attr, camera)
        return instance, data


class LaserSensorsVeh:
    def __init__(self):
        self.LEFT: Optional[Lidar] = None
        self.TOP: Optional[Lidar] = None
        self.RIGHT: Optional[Lidar] = None
        self.REAR: Optional[Lidar] = None

    def to_bytes(self):
        return b''.join(serialize(lidar) for lidar in [
            self.LEFT,
            self.TOP,
            self.RIGHT,
            self.REAR
        ])

    @classmethod
    def from_bytes(cls, data) -> 'LaserSensorsVeh':
        instance = cls()
        for attr in ['LEFT', 'TOP', 'RIGHT', 'REAR']:
            lidar, data = deserialize(data, Lidar)
            setattr(instance, attr, lidar)
        return instance, data


class VisionSensorsTow:
    def __init__(self):
        self.VIEW_1: Optional[Camera] = None
        self.VIEW_2: Optional[Camera] = None

    def to_bytes(self):
        return b''.join(serialize(camera) for camera in [
            self.VIEW_1,
            self.VIEW_2
        ])

    @classmethod
    def from_bytes(cls, data) -> 'VisionSensorsTow':
        instance = cls()
        for attr in ['VIEW_1', 'VIEW_2']:
            camera, data = deserialize(data, Camera)
            setattr(instance, attr, camera)
        return instance, data


class LaserSensorsTow:
    def __init__(self):
        self.VIEW_1: Optional[Lidar] = None
        self.VIEW_2: Optional[Lidar] = None
        self.UPPER_PLATFORM: Optional[Lidar] = None

    def to_bytes(self):
        return b''.join(serialize(lidar) for lidar in [
            self.VIEW_1,
            self.VIEW_2,
            self.UPPER_PLATFORM
        ])

    @classmethod
    def from_bytes(cls, data) -> 'LaserSensorsTow':
        instance = cls()
        for attr in ['VIEW_1', 'VIEW_2', 'UPPER_PLATFORM']:
            lidar, data = deserialize(data, Lidar)
            setattr(instance, attr, lidar)
        return instance, data


class Tower:
    def __init__(self):
        self.cameras: VisionSensorsTow = VisionSensorsTow()
        self.lidars: LaserSensorsTow = LaserSensorsTow()
        self.GNSS: Optional[GNSS] = GNSS()

    def to_bytes(self):
        tower_bytes = self.cameras.to_bytes() + self.lidars.to_bytes() + serialize(self.GNSS)
        return len(tower_bytes).to_bytes(INT_LENGTH, 'big') + tower_bytes

    @classmethod
    def from_bytes(cls, data) -> 'Tower':
        instance = cls()
        instance.cameras, data = VisionSensorsTow.from_bytes(data)
        instance.lidars, data = LaserSensorsTow.from_bytes(data)
        instance.GNSS, _ = deserialize(data, GNSS)
        return instance


class Vehicle:
    def __init__(self):
        self.cameras: VisionSensorsVeh = VisionSensorsVeh()
        self.lidars: LaserSensorsVeh = LaserSensorsVeh()
        self.IMU: IMU = IMU()
        self.GNSS: GNSS = GNSS()
        self.DYNAMICS: Dynamics = Dynamics()

    def to_bytes(self):
        vehicle_bytes = self.cameras.to_bytes() + self.lidars.to_bytes() + serialize(self.IMU) + serialize(
            self.GNSS) + serialize(self.DYNAMICS)
        return len(vehicle_bytes).to_bytes(INT_LENGTH, 'big') + vehicle_bytes

    @classmethod
    def from_bytes(cls, data) -> 'Vehicle':
        instance = cls()
        instance.cameras, data = VisionSensorsVeh.from_bytes(data)
        instance.lidars, data = LaserSensorsVeh.from_bytes(data)
        instance.IMU, data = deserialize(data, IMU)
        instance.GNSS, data = deserialize(data, GNSS)
        instance.DYNAMICS, _ = deserialize(data, Dynamics)
        return instance



import hashlib
import dill
from ameisedataset.miscellaneous import INT_LENGTH


def compute_checksum(data):
    # calculates the has value of a given bytestream - SHA256
    return hashlib.sha256(data).digest()


def read_data_block(data, dtype_length: int = INT_LENGTH):
    data_len = int.from_bytes(data[0:dtype_length], 'big')
    data_block_bytes = data[dtype_length:dtype_length + data_len]
    return data_block_bytes, data[dtype_length + data_len:]


def obj_to_bytes(obj) -> bytes:
    obj_bytes = dill.dumps(obj)
    obj_bytes_len = len(obj_bytes).to_bytes(INT_LENGTH, 'big')
    return obj_bytes_len + obj_bytes


def obj_from_bytes(data: bytes):
    return dill.loads(data)


def serialize(obj):
    if obj is None:
        return b'\x00\x00\x00\x00'
    obj_bytes = obj.to_bytes()
    obj_bytes_len = len(obj_bytes).to_bytes(INT_LENGTH, 'big')
    return obj_bytes_len + obj_bytes


def deserialize(data, cls, *args):
    obj_len = int.from_bytes(data[:INT_LENGTH], 'big')
    if obj_len == 0:
        return None, data[INT_LENGTH:]
    obj_data = data[INT_LENGTH:INT_LENGTH + obj_len]
    return cls.from_bytes(obj_data, *args), data[INT_LENGTH + obj_len:]



import numpy as np
from scipy.spatial.transform import Rotation as R
from typing import Tuple, List
from ameisedataset.data import CameraInformation, LidarInformation

def extract_translation_and_euler_from_matrix(mtx):
    # Extrahieren des Translationsvektors
    translation_vector = mtx[:3, 3]

    # Extrahieren der Rotationsmatrix und Umwandeln in Euler-Winkel (Radiant)
    rotation_matrix = mtx[:3, :3]
    rotation = R.from_matrix(rotation_matrix)
    euler_angles_rad = rotation.as_euler('xyz', degrees=False)

    return translation_vector, euler_angles_rad


def get_points_on_image(pcloud: List[np.ndarray], lidar_info: LidarInformation, cam_info: CameraInformation,
                        get_valid_only=True,
                        dtype_points_return=None) -> Tuple[np.array, List[Tuple]]:
    """Retrieve the projection matrix based on provided parameters."""
    if dtype_points_return is None:
        dtype_points_return = ['x', 'y', 'z', 'intensity', 'range']
    lidar_tf = Transformation('tbd', 'tbd', lidar_info.extrinsic.xyz, lidar_info.extrinsic.rpy)
    camera_tf = Transformation('tbd', 'tbd', cam_info.extrinsic.xyz, cam_info.extrinsic.rpy)
    camera_inverse_tf = camera_tf.invert_transformation()
    lidar_to_cam_tf = camera_inverse_tf.add_transformation(lidar_tf)
    rect_mtx = np.eye(4)
    rect_mtx[:3, :3] = cam_info.rectification_mtx
    proj_mtx = cam_info.projection_mtx

    projection = []
    points = []
    for point in pcloud:
        point_vals = np.array(point.tolist()[:3])
        # Transform points to new coordinate system
        point_in_camera = proj_mtx.dot(rect_mtx.dot(lidar_to_cam_tf.transformation_mtx.dot(np.append(point_vals[:3], 1))))
        # check if pts are behind the camera
        u = point_in_camera[0] / point_in_camera[2]
        v = point_in_camera[1] / point_in_camera[2]
        if get_valid_only:
            if point_in_camera[2] <= 0:
                continue
            elif 0 <= u < cam_info.shape[0] and 0 <= v < cam_info.shape[1]:
                projection.append((u, v))
                points.append(point[dtype_points_return])
            else:
                continue
        else:
            projection.append((u, v))
    return np.array(points, dtype=points[0].dtype), projection


class Transformation:
    def __init__(self, at, to, xyz, rpy):
        self._at = at
        self._to = to
        self._translation = xyz
        self._rotation = rpy
        self._update_transformation_matrix()

    @property
    def at(self):
        return self._at

    @at.setter
    def at(self, value):
        self._at = value

    @property
    def to(self):
        return self._to

    @to.setter
    def to(self, value):
        self._to = value

    @property
    def translation(self):
        return self._translation

    @translation.setter
    def translation(self, value):
        self._translation = np.array(value, dtype=float)
        self._update_transformation_matrix()

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, value):
        self._rotation = np.array(value, dtype=float)
        self._update_transformation_matrix()

    def _update_transformation_matrix(self):
        rotation = R.from_euler('xyz', self._rotation, degrees=False)
        rotation_matrix = rotation.as_matrix()
        self.transformation_mtx = np.identity(4)
        self.transformation_mtx[:3, :3] = rotation_matrix
        self.transformation_mtx[:3, 3] = self._translation

    def add_transformation(self, transformation_to_add):
        transformation_mtx_to_add = transformation_to_add.transformation_mtx
        new_transformation_mtx = np.dot(self.transformation_mtx, transformation_mtx_to_add)

        translation_vector, euler_angles = extract_translation_and_euler_from_matrix(new_transformation_mtx)

        new_transformation = Transformation(self.at, transformation_to_add.to, translation_vector, euler_angles)

        return new_transformation

    def invert_transformation(self):
        inverse_rotation_matrix = self.transformation_mtx[:3, :3].T
        inverse_translation_vector = -inverse_rotation_matrix @ self.transformation_mtx[:3, 3]
        inverse_transformation_matrix = np.identity(4)
        inverse_transformation_matrix[:3, :3] = inverse_rotation_matrix
        inverse_transformation_matrix[:3, 3] = inverse_translation_vector

        translation_vector, euler_angles = extract_translation_and_euler_from_matrix(inverse_transformation_matrix)

        inverse_transformation = Transformation(self.to, self.at, translation_vector, euler_angles)

        return inverse_transformation

    def __repr__(self):
        translation_str = ', '.join(f"{coord:.3f}" for coord in self.translation)
        rotation_str = ', '.join(f"{angle:.3f}" for angle in self.rotation)
        return (f"Transformation at {self._at} to {self._to},\n"
                f"  translation=[{translation_str}],\n"
                f"  rotation=[{rotation_str}]\n")


import os
import glob

from typing import List, Optional
from ameisedataset.data import *
from ameisedataset.miscellaneous import InvalidFileTypeError, obj_to_bytes, obj_from_bytes, INT_LENGTH


class DataRecord:
    def __init__(self, record_file: Optional[str] = None):
        # expect record_file to be an absolute readable path
        self.name: Optional[str] = record_file
        # TODO: store sequence name
        self.num_frames: int = 0
        self.frame_lengths: List[int] = []
        self.frames_data: bytes = b""
        if record_file is not None:
            if os.path.splitext(record_file)[1] != ".4mse":
                raise InvalidFileTypeError("This is not a valid AMEISE-Record file.")
            with open(record_file, 'rb') as file:
                """ TODO: implement checksum
                record_checksum = file.read(SHA256_CHECKSUM_LENGTH)
                if compute_checksum(combined_data) != record_checksum:
                    raise ChecksumError("Checksum mismatch. Data might be corrupted!")
                """
                # Read frame_lengths, array with num_frames entries (int)
                frame_lengths_len: int = int.from_bytes(file.read(INT_LENGTH), 'big')
                self.frame_lengths = obj_from_bytes(file.read(frame_lengths_len))
                # Read frames
                self.frames_data: bytes = file.read()
            self.num_frames: int = len(self.frame_lengths)
            self.name = os.path.splitext(os.path.basename(record_file))[0]

    def __len__(self):
        return self.num_frames

    def __getitem__(self, frame_index) -> Frame:
        if frame_index < 0 or frame_index >= len(self.frame_lengths):
            raise ValueError("Frame-Index out of range.")
        start_pos = sum(self.frame_lengths[:frame_index])
        end_pos = start_pos + self.frame_lengths[frame_index]
        return Frame.from_bytes(self.frames_data[start_pos:end_pos])

    @staticmethod
    def to_bytes(frames: List[Frame]) -> bytes:
        # frame lengths
        frame_lengths: List[int] = []
        # frame bytes
        frames_bytes = b""
        for _frame in frames:
            frame_bytes = _frame.to_bytes()
            frame_lengths.append(len(frame_bytes))
            frames_bytes += frame_bytes
        frame_lengths_bytes = obj_to_bytes(frame_lengths)
        # pack data sequence
        record_bytes = frame_lengths_bytes + frames_bytes
        # record_bytes_checksum = compute_checksum(record_bytes)
        return record_bytes  # record_bytes_checksum +


class Dataloader:
    # TODO: implement __iter__()
    def __init__(self, data_dir: str):
        self.data_dir: os.path = os.path.join(data_dir)
        self.record_map: List[str] = glob.glob(os.path.join(self.data_dir, '*.4mse'))

    def __len__(self):
        return len(self.record_map)

    def __getitem__(self, item) -> DataRecord:
        return DataRecord(record_file=self.record_map[item])

    def get_record_by_name(self, filename: str) -> Optional[DataRecord]:
        for record_path in self.record_map:
            if filename in record_path:
                return DataRecord(record_file=record_path)
        print(f"No record with name {filename} found.")
        return None

# amse_dataloader = Dataloader("/records")
# myRecord: DataRecord = amse_dataloader.get_record_by_name("2024-03-07-17-42-28_3_tower.4mse")
# myFrame: Frame = myRecord[7]
# print(myFrame.vehicle.cameras.FRONT_LEFT)



import numpy as np
from typing import Tuple, List

from ameisedataset.data import Pose, CameraInformation, LidarInformation, Image


def euler_to_rotation_matrix(roll, pitch, yaw):
    """Convert Euler angles to rotation matrix."""
    cos_r, sin_r = np.cos(roll), np.sin(roll)
    cos_p, sin_p = np.cos(pitch), np.sin(pitch)
    cos_y, sin_y = np.cos(yaw), np.sin(yaw)

    R_x = np.array([[1, 0, 0],
                    [0, cos_r, -sin_r],
                    [0, sin_r, cos_r]])

    R_y = np.array([[cos_p, 0, sin_p],
                    [0, 1, 0],
                    [-sin_p, 0, cos_p]])

    R_z = np.array([[cos_y, -sin_y, 0],
                    [sin_y, cos_y, 0],
                    [0, 0, 1]])

    R = np.dot(R_z, np.dot(R_y, R_x))
    return R


def create_transformation_matrix(translation, rotation):
    """Create a 4x4 homogeneous transformation matrix."""
    T = np.eye(4)
    T[:3, :3] = euler_to_rotation_matrix(rotation[0], rotation[1], rotation[2])
    T[:3, 3] = translation
    return T


def get_points_on_image(pcloud: List[np.ndarray], lidar_info: LidarInformation, cam_info: CameraInformation,
                        get_valid_only=True,
                        dtype_points_return=None) -> Tuple[np.array, List[Tuple]]:
    """Retrieve the projection matrix based on provided parameters."""
    if dtype_points_return is None:
        dtype_points_return = ['x', 'y', 'z', 'intensity', 'range']
    lidar_to_cam_tf_mtx = transform_to_sensor(lidar_info.extrinsic, cam_info.extrinsic)
    rect_mtx = np.eye(4)
    rect_mtx[:3, :3] = cam_info.rectification_mtx
    proj_mtx = cam_info.projection_mtx

    projection = []
    points = []
    for point in pcloud:
        point_vals = np.array(point.tolist()[:3])
        # Transform points to new coordinate system
        point_in_camera = proj_mtx.dot(rect_mtx.dot(lidar_to_cam_tf_mtx.dot(np.append(point_vals[:3], 1))))
        # check if pts are behind the camera
        u = point_in_camera[0] / point_in_camera[2]
        v = point_in_camera[1] / point_in_camera[2]
        if get_valid_only:
            if point_in_camera[2] <= 0:
                continue
            elif 0 <= u < cam_info.shape[0] and 0 <= v < cam_info.shape[1]:
                projection.append((u, v))
                points.append(point[dtype_points_return])
            else:
                continue
        else:
            projection.append((u, v))
    return np.array(points, dtype=points[0].dtype), projection


def transform_to_sensor(sensor1: Pose, sensor2: Pose):
    """Transform the data to the sensor's coordinate frame."""
    # Creating transformation matrices
    t1 = create_transformation_matrix(sensor1.xyz, sensor1.rpy)
    t2 = create_transformation_matrix(sensor2.xyz, sensor2.rpy)

    # Computing the transformation from the second sensor (new origin) to the first sensor
    t2_to_1 = np.dot(np.linalg.inv(t2), t1)
    return t2_to_1



import numpy as np
from enum import Enum
from typing import Tuple, Optional, Dict
from decimal import Decimal
from datetime import datetime, timedelta, timezone
from PIL import Image as PilImage
from ameisedataset.miscellaneous import read_data_block


def _convert_unix_to_utc(unix_timestamp_ns: Decimal, utc_offset_hours: int = 2) -> str:
    """
    Convert a Unix timestamp (in nanoseconds as Decimal) to a human-readable UTC string with a timezone offset.
    This function also displays milliseconds, microseconds, and nanoseconds.
    Parameters:
    - unix_timestamp_ns: Unix timestamp in nanoseconds as a Decimal.
    - offset_hours: UTC timezone offset in hours.
    Returns:
    - Human-readable UTC string with the given timezone offset and extended precision.
    """
    # Convert the Decimal to integer for calculations
    unix_timestamp_ns = int(unix_timestamp_ns)
    # Extract the whole seconds and the fractional part
    timestamp_s, fraction_ns = divmod(unix_timestamp_ns, int(1e9))
    milliseconds, remainder_ns = divmod(fraction_ns, int(1e6))
    microseconds, nanoseconds = divmod(remainder_ns, int(1e3))
    # Convert to datetime object and apply the offset
    dt = datetime.fromtimestamp(timestamp_s, timezone.utc) + timedelta(hours=utc_offset_hours)
    # Create the formatted string with extended precision
    formatted_time = dt.strftime('%Y-%m-%d %H:%M:%S')
    extended_precision = f".{milliseconds:03}{microseconds:03}{nanoseconds:03}"
    return formatted_time + extended_precision


class Velocity:
    def __init__(self, timestamp: Optional[Decimal] = None,
                 linear_velocity: Optional[np.array] = None,
                 angular_velocity: Optional[np.array] = None,
                 covariance: Optional[np.array] = None):
        self.timestamp: Optional[Decimal] = timestamp
        self.linear_velocity: Optional[np.array] = linear_velocity
        self.angular_velocity: Optional[np.array] = angular_velocity
        self.covariance: Optional[np.array] = covariance


class Motion:
    def __init__(self,
                 timestamp: Optional[Decimal] = None,
                 orientation: Optional[np.array] = None,
                 orientation_covariance: Optional[np.array] = None,
                 angular_velocity: Optional[np.array] = None,
                 angular_velocity_covariance: Optional[np.array] = None,
                 linear_acceleration: Optional[np.array] = None,
                 linear_acceleration_covariance: Optional[np.array] = None):
        self.timestamp: Optional[Decimal] = timestamp
        self.orientation: Optional[np.array] = orientation
        self.orientation_covariance: Optional[np.array] = orientation_covariance
        self.angular_velocity: Optional[np.array] = angular_velocity
        self.angular_velocity_covariance: Optional[np.array] = angular_velocity_covariance
        self.linear_acceleration: Optional[np.array] = linear_acceleration
        self.linear_acceleration_covariance: Optional[np.array] = linear_acceleration_covariance


class Position:
    class NavSatFixStatus(Enum):
        NO_FIX = -1
        FIX = 0
        SBAS_FIX = 1
        GBAS_FIX = 2

    class CovarianceType(Enum):
        UNKNOWN = 0
        APPROXIMATED = 1
        DIAGONAL_KNOWN = 2
        KNOWN = 3

    def __init__(self, timestamp: Optional[Decimal] = None, status: Optional[NavSatFixStatus] = None,
                 services: Optional[Dict[str, Optional[bool]]] = None, latitude: Optional[Decimal] = None,
                 longitude: Optional[Decimal] = None, altitude: Optional[Decimal] = None,
                 covariance: Optional[np.array] = None, covariance_type: Optional[CovarianceType] = None):
        self.timestamp: Optional[Decimal] = timestamp
        self.status = status
        self.services: Optional[Dict[str, Optional[bool]]] = self.init_services(services)
        self.latitude: Optional[Decimal] = latitude
        self.longitude: Optional[Decimal] = longitude
        self.altitude: Optional[Decimal] = altitude
        self.covariance: Optional[np.array] = covariance
        self.covariance_type = covariance_type

    def __iter__(self):
        return iter((self.latitude, self.longitude, self.timestamp))

    @staticmethod
    def init_services(services):
        default_services = {'GPS': None, 'Glonass': None, 'Galileo': None, 'Baidou': None}
        if services is None:
            return default_services
        for key in default_services:
            services.setdefault(key, default_services[key])
        return services


class Image:
    """
    Represents an image along with its metadata.
    Attributes:
        timestamp (Optional[Decimal]): Timestamp of the image as UNIX time, can be None.
        image (Optional[PilImage]): The actual image data, can be None.
    """

    def __init__(self, image: PilImage = None, timestamp: Optional[Decimal] = None):
        """
        Initializes the Image object with the provided image data and timestamp.
        Parameters:
            image (Optional[PilImage]): The actual image data. Defaults to None.
            timestamp (Optional[Decimal]): Timestamp of the image as UNIX time. Defaults to None.
        """
        self.image = image
        self.timestamp = timestamp

    def __getattr__(self, attr) -> PilImage:
        """
        Enables direct access to attributes of the `image` object.
        Parameters:
            attr (str): Name of the attribute to access.
        Returns:
            PilImage: Attribute value if it exists in the `image` object.
        Raises:
            AttributeError: If the attribute does not exist.
        """
        if hasattr(self.image, attr):
            return getattr(self.image, attr)
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    def to_bytes(self) -> bytes:
        encoded_img = self.image.tobytes()
        encoded_ts = str(self.timestamp).encode('utf-8')
        img_len = len(encoded_img).to_bytes(4, 'big')
        ts_len = len(encoded_ts).to_bytes(4, 'big')
        image_bytes = img_len + encoded_img + ts_len + encoded_ts
        return image_bytes

    @classmethod
    def from_bytes(cls, data: bytes, shape: Tuple[int, int]):
        img_bytes, data = read_data_block(data)
        ts_bytes, _ = read_data_block(data)
        img_instance = cls()
        img_instance.timestamp = Decimal(ts_bytes.decode('utf-8'))
        img_instance.image = PilImage.frombytes("RGB", shape, img_bytes)
        return img_instance

    def get_timestamp(self, utc=2):
        """
        Retrieves the UTC timestamp of the points.
        Args:
            utc (int, optional): Timezone offset in hours. Default is 2.
        Returns:
            str: The UTC timestamp of the points.
        """
        return _convert_unix_to_utc(self.timestamp, utc_offset_hours=utc)


class Points:
    """
    Represents a collection of points with an associated timestamp.
    Attributes:
        points (np.array): Array containing the points.
        timestamp (Decimal): Timestamp associated with the points.
    """

    def __init__(self, points: Optional[np.array] = None, timestamp: Optional[Decimal] = None):
        """
        Initializes the Points object with the provided points and timestamp.
        Parameters:
            points (np.array, optional): Array containing the points. Defaults to an empty array.
            timestamp (Decimal, optional): Timestamp associated with the points. Defaults to '0'.
        """
        self.points: np.array = points
        self.timestamp: Decimal = timestamp

    def __getattr__(self, attr) -> np.array:
        """
        Enables direct access to attributes of the `points` object.
        Parameters:
            attr (str): Name of the attribute to access.
        Returns:
            np.array: Attribute value if it exists.
        Raises:
            AttributeError: If the attribute does not exist.
        """
        if hasattr(self.points, attr):
            return getattr(self.points, attr)
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    def to_bytes(self) -> bytes:
        encoded_pts = self.points.tobytes()
        encoded_ts = str(self.timestamp).encode('utf-8')
        pts_len = len(encoded_pts).to_bytes(4, 'big')
        ts_len = len(encoded_ts).to_bytes(4, 'big')
        laser_bytes = pts_len + encoded_pts + ts_len + encoded_ts
        return laser_bytes

    @classmethod
    def from_bytes(cls, data: bytes, dtype: np.dtype):
        """
        Creates a Points instance from byte data.
        Parameters:
            data_bytes (bytes): Byte data representing the points.
            ts_data (bytes): Byte data representing the timestamp.
            dtype (np.dtype): Data type of the points.
        Returns:
            Points: A Points instance initialized with the provided data.
        """
        pts_bytes, data = read_data_block(data)
        ts_bytes, _ = read_data_block(data)
        pts_instance = cls()
        pts_instance.timestamp = Decimal(ts_bytes.decode('utf-8'))
        pts_instance.points = np.frombuffer(pts_bytes, dtype=dtype)
        return pts_instance

    def get_timestamp(self, utc=2):
        """
        Retrieves the UTC timestamp of the points.
        Args:
            utc (int, optional): Timezone offset in hours. Default is 2.
        Returns:
            str: The UTC timestamp of the points.
        """
        return _convert_unix_to_utc(self.timestamp, utc_offset_hours=utc)



from .error_exceptions import ChecksumError, InvalidFileTypeError
from .contstants import SHA256_CHECKSUM_LENGTH, INT_LENGTH, NUM_FRAMES_PER_RECORD
from .helper import compute_checksum, read_data_block, obj_from_bytes, obj_to_bytes, serialize, deserialize



from typing import Tuple, Optional
import numpy as np


class Pose:
    """
    Describes the position of a sensor in terms of its position and rotation relative to
    the reference coordinate system (Top_LiDAR).
    Attributes:
        xyz (Optional[np.array]): A 1x3 array representing the position of the sensor in the
                                  reference coordinate system (Top_LiDAR).
        rpy (Optional[np.array]): A 1x3 array representing the roll, pitch, and yaw angles of the sensor,
                                  describing its rotation in itself.
    """

    def __init__(self, xyz: Optional[np.array] = None, rpy: Optional[np.array] = None):
        """
        Initializes the Pose with default position and rotation as None.
        """
        self.xyz = xyz
        self.rpy = rpy


class TransformationMtx:
    """
    Represents a transformation matrix with separate rotation and translation components.
    Attributes:
        rotation (np.array): A 3x3 matrix representing the rotation component of the transformation.
        translation (np.array): A 1x3 matrix representing the translation component of the transformation.
    """

    def __init__(self, rotation: Optional[np.array] = None, translation: Optional[np.array] = None):
        """
        Initializes the TransformationMtx with zero rotation and translation matrices.
        """
        self.rotation = rotation
        self.translation = translation


class ROI:
    """
    Represents a Region of Interest (ROI) defined by its offset and dimensions.
    Attributes:
        x_offset (Optional[int]): The horizontal offset of the ROI.
        y_offset (Optional[int]): The vertical offset of the ROI.
        width (Optional[int]): The width of the ROI.
        height (Optional[int]): The height of the ROI.
    """

    def __init__(self, x_off: Optional[int] = None, y_off: Optional[int] = None, width: Optional[int] = None,
                 height: Optional[int] = None):
        """
        Initializes the ROI with the provided offset and dimensions.
        Parameters:
            x_off (Optional[int], optional): Horizontal offset. Defaults to None.
            y_off (Optional[int], optional): Vertical offset. Defaults to None.
            width (Optional[int], optional): Width of the ROI. Defaults to None.
            height (Optional[int], optional): Height of the ROI. Defaults to None.
        """
        self.x_offset = x_off
        self.y_offset = y_off
        self.width = width
        self.height = height

    def __iter__(self):
        """
        Allows iteration over the ROI attributes in the order: x_offset, y_offset, width, height.
        Returns:
            iterator: An iterator over the ROI attributes.
        """
        return iter((self.x_offset, self.y_offset, self.width, self.height))


class DynamicsInformation:
    def __init__(self, velocity_source: Optional[str] = None):
        self.velocity_source = velocity_source


class IMUInformation:
    def __init__(self, model_name: Optional[str] = None, extrinsic: Optional[Pose] = None):
        self.model_name = model_name
        self.extrinsic = extrinsic


class GNSSInformation:
    def __init__(self, model_name: Optional[str] = None, extrinsic: Optional[Pose] = None):
        self.model_name = model_name
        self.extrinsic = extrinsic


class CameraInformation:

    def __init__(self,
                 name: str,
                 model_name: Optional[str] = None,
                 shape: Optional[Tuple[int, int]] = None,
                 distortion_type: Optional[str] = None,
                 camera_mtx: Optional[np.array] = None,
                 distortion_mtx: Optional[np.array] = None,
                 rectification_mtx: Optional[np.array] = None,
                 projection_mtx: Optional[np.array] = None,
                 region_of_interest: Optional[ROI] = None,
                 camera_type: Optional[str] = None,
                 focal_length: Optional[int] = None,
                 aperture: Optional[int] = None,
                 exposure_time: Optional[int] = None,
                 extrinsic: Optional[Pose] = None,
                 stereo_transform: Optional[TransformationMtx] = None):
        self.name = name
        self.model_name = model_name
        self.shape = shape
        self.distortion_type = distortion_type
        self.camera_mtx = camera_mtx
        self.distortion_mtx = distortion_mtx
        self.rectification_mtx = rectification_mtx
        self.projection_mtx = projection_mtx
        self.region_of_interest = region_of_interest
        self.camera_type = camera_type
        self.focal_length = focal_length
        self.aperture = aperture
        self.exposure_time = exposure_time
        self.extrinsic = extrinsic
        self.stereo_transform = stereo_transform


class LidarInformation:
    ouster_datatype_structure = {
        'names': [
            'x',  # x-coordinate of the point
            'y',  # y-coordinate of the point
            'z',  # z-coordinate of the point
            'intensity',  # Intensity of the point
            't',  # Time after the frame timestamp in ns
            'reflectivity',  # Reflectivity of the point
            'ring',  # Ring number (for multi-beam LiDARs)
            'ambient',  # Ambient light intensity
            'range'  # Distance from the LiDAR sensor to the measured point (hypotenuse) in mm.
        ],
        'formats': ['<f4', '<f4', '<f4', '<f4', '<u4', '<u2', '<u2', '<u2', '<u4'],
        'offsets': [0, 4, 8, 16, 20, 24, 26, 28, 32],
        'itemsize': 48
    }

    blickfeld_datatype_structure = {
        'names': [
            'x',  # x-coordinate of the point
            'y',  # y-coordinate of the point
            'z',  # z-coordinate of the point
            'intensity',  # Intensity of the point
            'point_id',
        ],
        'formats': ['<f4', '<f4', '<f4', '<u4', '<u4'],
        'offsets': [0, 4, 8, 12, 16],
        'itemsize': 20
    }

    def __init__(self,
                 name: str,
                 model_name: Optional[str] = None,
                 beam_altitude_angles: Optional[np.array] = None,
                 beam_azimuth_angles: Optional[np.array] = None,
                 lidar_origin_to_beam_origin_mm: Optional[np.array] = None,
                 columns_per_frame: Optional[int] = None,
                 pixels_per_column: Optional[int] = None,
                 phase_lock_offset: Optional[int] = None,
                 lidar_to_sensor_transform: Optional[np.array] = None,
                 extrinsic: Optional[Pose] = None):

        self._model_name = None
        self.name = name
        self.model_name = model_name  # This will trigger the setter and update dtype
        self.beam_altitude_angles = beam_altitude_angles
        self.beam_azimuth_angles = beam_azimuth_angles
        self.lidar_origin_to_beam_origin_mm = lidar_origin_to_beam_origin_mm
        self.columns_per_frame = columns_per_frame
        self.pixels_per_column = pixels_per_column
        self.phase_lock_offset = phase_lock_offset
        self.lidar_to_sensor_transform = lidar_to_sensor_transform
        self.extrinsic = extrinsic

    @property
    def model_name(self) -> Optional[str]:
        return self._model_name

    @model_name.setter
    def model_name(self, value: Optional[str]):
        self._model_name = value
        if value is not None:
            if 'OS' in value:
                dtype_structure = LidarInformation.ouster_datatype_structure
            elif 'Blickfeld' in value:
                dtype_structure = LidarInformation.blickfeld_datatype_structure
            else:
                dtype_structure = None
            self.dtype = np.dtype(dtype_structure) if dtype_structure else None
        else:
            self.dtype = None



from PIL import ImageDraw, Image
import numpy as np
import matplotlib


def show_disparity_map(disparity_map, cmap_name="viridis", val_min=None, val_max=None):
    cmap = matplotlib.colormaps[cmap_name]
    val_min = val_min if val_min is not None else np.min(disparity_map)
    val_max = val_max if val_max is not None else np.max(disparity_map)
    mask = disparity_map > 10 * val_max
    norm_values = np.where(mask, 0, (disparity_map - val_min) / (val_max - val_min))
    colored_map = cmap(norm_values)
    colored_map[mask] = [0, 0, 0, 1]  # Set masked values to black
    colored_map = (colored_map[:, :, :3] * 255).astype(np.uint8)
    img = Image.fromarray(colored_map)
    return img


def plot_points_on_image(img, points, values, cmap_name="inferno", radius=2, val_min=None, val_max=None):
    draw = ImageDraw.Draw(img)
    cmap = matplotlib.colormaps[cmap_name + "_r"]
    val_min = val_min * 1000 if val_min is not None else np.min(values)
    val_max = val_max * 1000 if val_max is not None else np.max(values)
    norm_values = (values - val_min) / (val_max - val_min)
    for punkt, value in zip(points, norm_values):
        x, y = punkt
        rgba = cmap(value)
        farbe = (int(rgba[0] * 255), int(rgba[1] * 255), int(rgba[2] * 255))
        draw.ellipse([(x - radius, y - radius), (x + radius, y + radius)], fill=farbe)
    return img



from .metadata import CameraInformation, LidarInformation, IMUInformation, GNSSInformation, Pose, DynamicsInformation, \
    ROI, TransformationMtx
from .data import Image, Points, Position, Motion, Velocity
from .sensors import Lidar, Camera, IMU, GNSS, Dynamics
from .agent import Tower, Vehicle, VisionSensorsVeh, VisionSensorsTow, LaserSensorsVeh, LaserSensorsTow
from .frame import Frame



import numpy as np
from typing import List, Optional
from ameisedataset.miscellaneous import serialize, deserialize, obj_to_bytes, obj_from_bytes, read_data_block
from ameisedataset.data import Image, Points, Motion, Position, CameraInformation, LidarInformation, GNSSInformation, \
    IMUInformation, Velocity, DynamicsInformation


class Camera:
    def __init__(self, info: Optional[CameraInformation] = None, image: Optional[Image] = None):
        self.info = info
        self.image = image

    def __getattr__(self, attr) -> np.array:
        if hasattr(self.image, attr):
            return getattr(self.image, attr)
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    def to_bytes(self) -> bytes:
        return obj_to_bytes(self.info) + serialize(self.image)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'Camera':
        instance = cls()
        info_bytes, data = read_data_block(data)
        setattr(instance, 'info', obj_from_bytes(info_bytes))
        image, _ = deserialize(data, Image, instance.info.shape)
        setattr(instance, 'image', image)
        return instance


class Lidar:
    def __init__(self, info: Optional[LidarInformation] = None, points: Optional[Points] = None):
        self.info = info
        self.points = points

    def __getattr__(self, attr) -> np.array:
        if hasattr(self.points, attr):
            return getattr(self.points, attr)
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    def to_bytes(self) -> bytes:
        return obj_to_bytes(self.info) + serialize(self.points)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'Lidar':
        instance = cls()
        info_bytes, data = read_data_block(data)
        setattr(instance, 'info', obj_from_bytes(info_bytes))
        points, _ = deserialize(data, Points, instance.info.dtype)
        setattr(instance, 'points', points)
        return instance


class IMU:
    def __init__(self, info: Optional[IMUInformation] = None):
        self.info = info
        self.motion: List[Motion] = []

    def __getattr__(self, attr) -> np.array:
        if hasattr(self.ekf, attr):
            return getattr(self.ekf, attr)
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    def to_bytes(self) -> bytes:
        return obj_to_bytes(self.info) + obj_to_bytes(self.motion)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'IMU':
        instance = cls()
        info_bytes, data = read_data_block(data)
        setattr(instance, 'info', obj_from_bytes(info_bytes))
        motion_bytes, _ = read_data_block(data)
        setattr(instance, 'motion', obj_from_bytes(motion_bytes))
        return instance


class Dynamics:
    def __init__(self, info: Optional[DynamicsInformation] = None):
        self.info = info
        self.velocity: List[Velocity] = []

    def to_bytes(self) -> bytes:
        return obj_to_bytes(self.info) + obj_to_bytes(self.velocity)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'Dynamics':
        instance = cls()
        info_bytes, data = read_data_block(data)
        setattr(instance, 'info', obj_from_bytes(info_bytes))
        velocity_bytes, _ = read_data_block(data)
        setattr(instance, 'velocity', obj_from_bytes(velocity_bytes))
        return instance


class GNSS:
    def __init__(self, info: Optional[GNSSInformation] = None):
        self.info = info
        self.position: List[Position] = []

    def __getattr__(self, attr) -> np.array:
        if hasattr(self.ekf, attr):
            return getattr(self.ekf, attr)
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    def to_bytes(self) -> bytes:
        return obj_to_bytes(self.info) + obj_to_bytes(self.position)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'GNSS':
        instance = cls()
        info_bytes, data = read_data_block(data)
        setattr(instance, 'info', obj_from_bytes(info_bytes))
        position_bytes, _ = read_data_block(data)
        setattr(instance, 'position', obj_from_bytes(position_bytes))
        return instance
